"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MidwayContainer = void 0;
const decorator_1 = require("@midwayjs/decorator");
const injection_1 = require("injection");
const is = require("is-type-of");
const constant_1 = require("./constant");
const glob_1 = require("@midwayjs/glob");
const graphviz = require('graphviz');
const camelcase = require('camelcase');
const debug = require('debug')('midway:container');
const CONTROLLERS = 'controllers';
const MIDDLEWARES = 'middlewares';
const TYPE_LOGGER = 'logger';
const TYPE_PLUGIN = 'plugin';
class BaseParser {
    constructor(container) {
        this.container = container;
    }
}
/**
 * 用于xml解析扩展
 * <controllers />
 */
class ControllerDefinitionParser extends BaseParser {
    constructor() {
        super(...arguments);
        this.name = CONTROLLERS;
    }
    parse(ele, context) {
        const definition = new injection_1.XmlObjectDefinition(ele);
        context.parser.parseElementNodes(definition, ele, context);
        this.container.controllersIds.push(definition.id);
        return definition;
    }
}
/**
 * 用于xml解析扩展
 * <middlewares />
 */
class MiddlewareDefinitionParser extends BaseParser {
    constructor() {
        super(...arguments);
        this.name = MIDDLEWARES;
    }
    parse(ele, context) {
        const definition = new injection_1.XmlObjectDefinition(ele);
        context.parser.parseElementNodes(definition, ele, context);
        this.container.middlewaresIds.push(definition.id);
        return definition;
    }
}
class ManagedLogger {
    constructor() {
        this.type = TYPE_LOGGER;
    }
}
class LoggerParser {
    get name() {
        return TYPE_LOGGER;
    }
    parse(ele, context) {
        const log = new ManagedLogger();
        log.name = ele.getAttribute('name').trim();
        return log;
    }
}
class LoggerResolver {
    constructor(container) {
        this.container = container;
    }
    get type() {
        return TYPE_LOGGER;
    }
    resolve(managed) {
        const log = managed;
        if (log.name) {
            return this.container.findHandlerHook(constant_1.MidwayHandlerKey.LOGGER)(log.name);
        }
        return this.container.findHandlerHook(constant_1.MidwayHandlerKey.LOGGER)(log.type);
    }
    async resolveAsync(managed) {
        return this.resolve(managed);
    }
}
class ManagedPlugin {
    constructor() {
        this.type = TYPE_PLUGIN;
    }
}
class PluginParser {
    get name() {
        return TYPE_PLUGIN;
    }
    parse(ele, context) {
        const plugin = new ManagedPlugin();
        plugin.name = ele.getAttribute('name').trim();
        return plugin;
    }
}
class PluginResolver {
    constructor(container) {
        this.container = container;
    }
    get type() {
        return TYPE_PLUGIN;
    }
    resolve(managed) {
        const p = managed;
        return this.container.findHandlerHook(constant_1.MidwayHandlerKey.PLUGIN)(p.name);
    }
    async resolveAsync(managed) {
        return this.resolve(managed);
    }
}
class MidwayContainer extends injection_1.Container {
    constructor(baseDir = process.cwd(), parent = undefined, isTsMode = true) {
        super(baseDir, parent);
        this.controllersIds = [];
        this.middlewaresIds = [];
        // 仅仅用于兼容requestContainer的ctx
        this.ctx = {};
        this.isTsMode = isTsMode;
    }
    init() {
        this.handlerMap = new Map();
        super.init();
        if (!this.isTsMode) {
            // xml扩展 <logger name=""/> <plugin name="hsfclient"/>
            this.parser.objectElementParser.registerParser(new LoggerParser());
            this.getManagedResolverFactory().registerResolver(new LoggerResolver(this));
            this.parser.objectElementParser.registerParser(new PluginParser());
            this.getManagedResolverFactory().registerResolver(new PluginResolver(this));
            this.parser.registerParser(new ControllerDefinitionParser(this));
            this.parser.registerParser(new MiddlewareDefinitionParser(this));
        }
        this.registerEachCreatedHook();
        // 防止直接从applicationContext.getAsync or get对象实例时依赖当前上下文信息出错
        // ctx is in requestContainer
        this.registerObject('ctx', this.ctx);
    }
    /**
     * update current context in applicationContext
     * for mock and other case
     * @param ctx ctx
     * @deprecated
     */
    updateContext(ctx) {
        this.ctx = Object.assign({}, ctx || {});
    }
    /**
     * load directory and traverse file to find bind class
     * @param opts
     */
    load(opts) {
        const loadDirs = [].concat(opts.loadDir || []);
        // TODO set 去重
        for (const dir of loadDirs) {
            const fileResults = glob_1.run(['**/**.ts', '**/**.tsx', '**/**.js'].concat(opts.pattern || []), {
                cwd: dir,
                ignore: [
                    '**/**.d.ts',
                    '**/logs/**',
                    '**/run/**',
                    '**/public/**',
                    '**/view/**',
                    '**/views/**'
                ].concat(opts.ignore || []),
            });
            for (const file of fileResults) {
                debug(`binding file => ${file}`);
                const exports = require(file);
                this.bindClass(exports);
            }
        }
    }
    bindClass(exports) {
        if (is.class(exports) || is.function(exports)) {
            this.bindModule(exports);
        }
        else {
            for (const m in exports) {
                const module = exports[m];
                if (is.class(module) || is.function(module)) {
                    this.bindModule(module);
                }
            }
        }
    }
    bindModule(module) {
        if (is.class(module)) {
            const providerId = injection_1.getProviderId(module);
            if (providerId) {
                this.bind(providerId, module);
            }
            else {
                if (!this.isTsMode) {
                    // inject by name in js
                    this.bind(camelcase(module.name), module);
                }
            }
        }
        else {
            const info = module[constant_1.FUNCTION_INJECT_KEY];
            if (info && info.id) {
                if (!info.scope) {
                    info.scope = injection_1.ScopeEnum.Request;
                }
                this.bind(info.id, module, {
                    scope: info.scope,
                    isAutowire: info.isAutowire
                });
            }
        }
    }
    createChild() {
        const child = new injection_1.Container();
        child.parent = this;
        return child;
    }
    registerEachCreatedHook() {
        // register constructor inject
        this.beforeEachCreated((target, constructorArgs, context) => {
            let constructorMetaData;
            try {
                constructorMetaData = injection_1.getClassMetadata(decorator_1.CLASS_KEY_CONSTRUCTOR, target);
            }
            catch (e) {
                debug(`beforeEachCreated error ${e.stack}`);
            }
            // lack of field
            if (constructorMetaData && constructorArgs) {
                for (const idx in constructorMetaData) {
                    const index = parseInt(idx, 10);
                    const propertyMeta = constructorMetaData[index];
                    let result;
                    switch (propertyMeta.type) {
                        case 'config':
                            result = this.findHandlerHook(constant_1.MidwayHandlerKey.CONFIG)(propertyMeta.key);
                            break;
                        case 'logger':
                            result = this.findHandlerHook(constant_1.MidwayHandlerKey.LOGGER)(propertyMeta.key);
                            break;
                        case 'plugin':
                            result = this.findHandlerHook(constant_1.MidwayHandlerKey.PLUGIN)(propertyMeta.key);
                            break;
                    }
                    constructorArgs[index] = result;
                }
            }
        });
        // register property inject
        this.afterEachCreated((instance, context, definition) => {
            // 处理配置装饰器
            const configSetterProps = injection_1.getClassMetadata(decorator_1.CONFIG_KEY, instance);
            this.defineGetterPropertyValue(configSetterProps, instance, this.findHandlerHook(constant_1.MidwayHandlerKey.CONFIG));
            // 处理插件装饰器
            const pluginSetterProps = injection_1.getClassMetadata(decorator_1.PLUGIN_KEY, instance);
            this.defineGetterPropertyValue(pluginSetterProps, instance, this.findHandlerHook(constant_1.MidwayHandlerKey.PLUGIN));
            // 处理日志装饰器
            const loggerSetterProps = injection_1.getClassMetadata(decorator_1.LOGGER_KEY, instance);
            this.defineGetterPropertyValue(loggerSetterProps, instance, this.findHandlerHook(constant_1.MidwayHandlerKey.LOGGER));
            // 表示非ts annotation模式
            if (!this.isTsMode && !pluginSetterProps && !loggerSetterProps && definition.isAutowire()) {
                // this.$$xxx = null; 用来注入config
                // this.$xxx = null; 用来注入 logger 或者 插件
                injection_1.Autowire.patchDollar(instance, context, (key) => {
                    if (key[0] === '$') {
                        return this.findHandlerHook(constant_1.MidwayHandlerKey.CONFIG)(key.slice(1));
                    }
                    try {
                        const v = this.findHandlerHook(constant_1.MidwayHandlerKey.PLUGIN)(key);
                        if (v) {
                            return v;
                        }
                    }
                    catch (e) {
                    }
                    return this.findHandlerHook(constant_1.MidwayHandlerKey.LOGGER)(key);
                });
            }
        });
    }
    /**
     * binding getter method for decorator
     *
     * @param setterProps
     * @param instance
     * @param getterHandler
     */
    defineGetterPropertyValue(setterProps, instance, getterHandler) {
        if (setterProps && getterHandler) {
            for (const prop of setterProps) {
                if (prop.propertyName) {
                    Object.defineProperty(instance, prop.propertyName, {
                        get: () => getterHandler(prop.key, instance),
                        configurable: false,
                        enumerable: true
                    });
                }
            }
        }
    }
    registerDataHandler(handlerType, handler) {
        this.handlerMap.set(handlerType, handler);
    }
    registerCustomBinding(objectDefinition, target) {
        super.registerCustomBinding(objectDefinition, target);
        // Override the default scope to request
        const objDefOptions = injection_1.getObjectDefinition(target);
        if (objDefOptions && !objDefOptions.scope) {
            debug(`register @scope to default value(request), id=${objectDefinition.id}`);
            objectDefinition.scope = injection_1.ScopeEnum.Request;
        }
    }
    dumpDependency() {
        const g = graphviz.digraph('G');
        for (const [id, module] of this.dependencyMap.entries()) {
            g.addNode(id, { label: `${id}(${module.name})\nscope:${module.scope}`, fontsize: '10' });
            module.properties.forEach((depId) => {
                g.addEdge(id, depId, { label: `properties`, fontsize: '8' });
            });
            module.constructorArgs.forEach((depId) => {
                g.addEdge(id, depId, { label: 'constructor', fontsize: '8' });
            });
        }
        try {
            return g.to_dot();
        }
        catch (err) {
            console.error('generate injection dependency tree fail, err = ', err.message);
        }
    }
    /**
     * get hook from current map or parent map
     * @param hookKey
     */
    findHandlerHook(hookKey) {
        if (this.handlerMap.has(hookKey)) {
            return this.handlerMap.get(hookKey);
        }
        if (this.parent) {
            return this.parent.findHandlerHook(hookKey);
        }
    }
}
exports.MidwayContainer = MidwayContainer;
//# sourceMappingURL=container.js.map